! KILO LISP 22, a kilo byte-sized LISP system
! Nils M Holm, 2019, 2020
! In the public domain

!!! Version 22 is not compatible to the original version.
!!! See the CHANGES file and manual for details.

! Internal differences to the original C version:
! - Error handling is different, because there is no longjmp()
! - Console I/O is direct, see READCON
! - "KL -" will read SYSIN instead of the console

module klisp(t3x, ttyctl);

object	t[t3x], tty[ttyctl];

const	NNODES	= 12288;

const	SYMLEN	= 64;
const	BUFLEN	= 128;
const	EVDEPTH	= 512;
const	PRDEPTH	= 64;
const	NLOAD	= 3;

const	SPCL	= NNODES;
const	NIL	= NNODES+1;
const	EOT	= NNODES+2;
const	DOT	= NNODES+3;
const	RPAREN	= NNODES+4;
const	UNDEF	= NNODES+5;

strlen(s) return t.memscan(s, 0, 32767);
strequ(a, b) return t.memcomp(a, b, strlen(a)+1) = 0;
strcopy(a, b) return t.memcopy(a, b, strlen(b)+1);

var Ntoab::20;

ntoa(n) do var q, p;
	p := @Ntoab::19;
	q := p;
	p::0 := 0;
	while (n \/ p = q) do
		p := p-1;
		p::0 := n mod 10 + '0';
		n := n / 10;
	end
	return p;
end

var	Xpos, Ypos;

clear() do
	tty.clear();
	Xpos := 0;
	Ypos := 0;
end

var	Pass;

pr(s) do var i;
	ie (Pass) do
		t.write(T3X.SYSOUT, s, strlen(s));
	end
	else do
		i := 0;
		while (s::i) do
			ie (s::i = '\n') do
				Ypos := Ypos+ 1;
				tty.move(Xpos, Ypos);
			end
			else ie (s::i = '\r') do
				Xpos := 0;
				tty.move(Xpos, Ypos);
			end
			else do
				tty.writec(s::i);
				Xpos := Xpos + 1;
			end
			if (Xpos >= tty.columns()) do
				Xpos := 0;
				Ypos := Ypos + 1;
				tty.move(Xpos, Ypos);
			end
			if (Ypos >= tty.lines()-1) do
				tty.scroll(0, tty.lines()-2);
				Ypos := tty.lines() - 2;
				tty.move(Xpos, Ypos);
			end
			i := i+1;
		end
	end
end

nl() do var b::3;
	ie (Pass)
		pr(t.newline(b));
	else
		pr("\r\n");
end

var	Input, Inp, Ink;
var	Inbuf, Buffer::BUFLEN;
var	Rejected;

flushinp() do
	Input := T3X.SYSIN;
	Inbuf := Buffer;
	Inp := 0;
	Ink := 0;
	Rejected := EOT;
end

var	Auxb1::BUFLEN, Auxb2::BUFLEN;

rotatebufs(b) do var i, c;
	for (i=0, BUFLEN) do
		c := Auxb1::i;
		Auxb1::i := Auxb2::i;
		Auxb2::i := b::i;
		b::i := c;
	end
end

decl error(2);

readcon(b) do var p0, p, r, m, k, n;
	m := tty.columns() - 4;
	p0 := Xpos;
	p := 0;
	r := 1;
	n := 0;
	b::0 := 0;
	while (1) do
		if (r) do
			tty.move(p0, Ypos);
			tty.writes(b);
			tty.writec('\s');
			r := 0;
		end
		tty.move(p+p0, Ypos);
		k := tty.readc();
		ie (k = 'B'-'@' \/ k = ttyctl.K_LEFT) do
			if (p > 0) p := p-1;
		end
		else ie (k = 'F'-'@' \/ k = ttyctl.K_RIGHT) do
			if (p < n) p := p+1;
		end
		else ie (k = 'A'-'@') do
			p := 0;
		end
		else ie (k = 'E'-'@') do
			p := n;
		end
		else ie (k = 'L'-'@') do
			clear();
			if (p0) pr("* ");
			r := 1;
		end
		else ie (k = 'P'-'@' \/ k = ttyctl.K_UP) do
			rotatebufs(b);
			n := strlen(b);
			p := n;
			r := 1;
			tty.move(Xpos, Ypos);
			tty.clreol();
		end
		else ie (k = '\r' \/ k = '\n') do
			if (n > 0) do
				if (Auxb1::0 \= 0)
					strcopy(Auxb2, Auxb1);
				strcopy(Auxb1, b);
			end
			b::n := k;
			n := n+1;
			b::n := 0;
			nl();
			return n;
		end
		else ie (k = 'C'-'@') do
			nl();
			b::0 := 0;
			error("abort", UNDEF);
			return 0;
		end
		else ie (k = 'U'-'@') do
			b::0 := 0;
			n := 0;
			p := 0;
			r := 1;
			tty.move(p0, Ypos);
			tty.clreol();
		end
		else ie ((k = '\b' \/ k = ttyctl.K_BKSP \/ k = 127) /\ p > 0)
		do
			n := n-1;
			p := p-1;
			t.memcopy(@b::p, @b::(p+1), n-p);
			b::n := 0;
			r := 1;
		end
		else ie (k = 4 /\ n = 0) do
			b::n := 4;
			b::(n+1) := 0;
			pr("^D");
			nl();
			return n;
		end
		else if ('\s' <= k /\ k <= '~') do
			if (n >= m) loop;
			t.memcopy(@b::(p+1), @b::p, n-p);
			b::p := k;
			p := p+1;
			n := n+1;
			b::n := 0;
			r := 1;
		end
	end
end

rdch() do var c;
	if (Rejected \= EOT) do
		c := Rejected;
		Rejected := EOT;
		return c;
	end
	if (Inp >= Ink) do
		ie (Input = T3X.SYSIN /\ \Pass)
			Ink := readcon(Inbuf);
		else
			Ink := t.read(Input, Inbuf, BUFLEN);
		if (Ink < 1) return EOT;
		Inp := 0;
	end
	c := Inbuf::Inp;
	Inp := Inp+1;
	if (c = 4) return EOT;
	return c;
end

rdchci() do var c;
	c := rdch();
	if (c >= 'A' /\ c <= 'Z') return c+32;
	return c;
end

message(s) do
	if (Pass) return;
	tty.move(0, tty.lines()-1);
	tty.clreol();
	tty.writes(s);
	tty.move(Xpos, Ypos);
end

fini(x) do
	if (\Pass) do
		tty.reset();
		tty.mode(0);
		tty.fini();
	end
	ie (x) halt 1; else halt 0;
end

var	Errflag;

decl print(1);

error(m, n) do
	if (Errflag) return;
	pr("? ");
	pr(m);
	if (n \= UNDEF) do
		pr(": ");
		print(n);
	end
	nl();
	flushinp();
	Errflag := %1;
	if (Pass) fini(1);
	return NIL;
end

fatal(m) do
	message("");
	error(m, UNDEF);
	pr("? aborting");
	nl();
	fini(1);
end

var	Car[NNODES],
	Cdr[NNODES];

var	Tag::NNODES;

const	ATOMB	= 0x01;
const	MARKB	= 0x02;
const	SWAPB	= 0x04;

var	Freelist;

atomp(n) return n >= SPCL \/ Car[n] < SPCL /\ Tag::Car[n] & ATOMB;

symbolp(n) return n < SPCL /\ Car[n] < SPCL /\ Tag::Car[n] & ATOMB;

! Deutsch/Schorr/Waite graph marker

mark(n) do var p, x;
	p := NIL;
	while (1) do
		ie (n >= SPCL \/ Tag::n & MARKB) do
			if (p = NIL) leave;
			ie (Tag::p & SWAPB) do
				x := Cdr[p];
				Cdr[p] := Car[p];
				Car[p] := n;
				Tag::p := Tag::p & ~SWAPB;
				n := x;
			end
			else do
				x := p;
				p := Cdr[x];
				Cdr[x] := n;
				n := x;
			end
		end
		else ie (Tag::n & ATOMB) do
			x := Cdr[n];
			Cdr[n] := p;
			p := n;
			n := x;
			Tag::p := Tag::p | MARKB;
		end
		else do
			x := Car[n];
			Car[n] := p;
			Tag::n := Tag::n | MARKB;
			p := n;
			n := x;
			Tag::p := Tag::p | SWAPB;
		end
	end
end

var	Symbols;
var	Acc, Env;
var	Stack, Mstack;
var	Tmpcar, Tmpcdr;
var	Znode;

var	Verbose_GC;

gc() do var i, k;
	if (Verbose_GC) pr("GC: ");
	mark(Znode);
	mark(Acc);
	mark(Env);
	mark(Symbols);
	mark(Stack);
	mark(Mstack);
	mark(Tmpcar);
	mark(Tmpcdr);
	k := 0;
	Freelist := NIL;
	for (i=0, NNODES) do
		ie (Tag::i & MARKB) do
			Tag::i := Tag::i & ~MARKB;
		end
		else do
			Cdr[i] := Freelist;
			Freelist := i;
			k := k+1;
		end
	end
	if (Verbose_GC) do
		pr(ntoa(k));
		pr(" NODES");
		nl();
	end
	return k;
end

cons3(a, d, ta) do var n;
	if (Freelist = NIL) do
		Tmpcdr := d;
		if (\ta) Tmpcar := a;
		gc();
		Tmpcar := NIL;
		Tmpcdr := NIL;
		if (Freelist = NIL) do
			error("out of nodes", UNDEF);
			return Znode;
		end
	end
	n := Freelist;
	Freelist := Cdr[Freelist];
	Car[n] := a;
	Cdr[n] := d;
	Tag::n := ta;
	return n;
end

cons(a, d) return cons3(a, d, 0);

caar(x) return Car[Car[x]];
cadr(x) return Car[Cdr[x]];
cdar(x) return Cdr[Car[x]];
cddr(x) return Cdr[Cdr[x]];

caadr(x) return Car[Car[Cdr[x]]];
cadar(x) return Car[Cdr[Car[x]]];
caddr(x) return Car[Cdr[Cdr[x]]];
cdadr(x) return Cdr[Car[Cdr[x]]];
cddar(x) return Cdr[Cdr[Car[x]]];

caddar(x) return Car[Cdr[Cdr[Car[x]]]];
cdddar(x) return Cdr[Cdr[Cdr[Car[x]]]];

nrev(n) do var m, h;
	m := NIL;
	while (n \= NIL) do
		h := Cdr[n];
		Cdr[n] := m;
		m := n;
		n := h;
	end
	return m;
end

save(n) Stack := cons(n, Stack);

unsave(k) do var n;
	while (k) do
		if (Stack = NIL) fatal("stack empty");
		n := Car[Stack];
		Stack := Cdr[Stack];
		k := k-1;
	end
	return n;
end

msave(n) Mstack := cons3(n, Mstack, ATOMB);

munsave() do var n;
	if (Mstack = NIL) fatal("mstack empty");
	n := Car[Mstack];
	Mstack := Cdr[Mstack];
	return n;
end

strsym(s) do var i, n;
        i := 0;
        if (s::i = 0) return NIL;
	n := cons3(NIL, NIL, ATOMB);
        save(n);
        while (1) do
                Car[n] := s::i << 8 | s::(i+1);
                if (s::(i+1) = 0 \/ s::(i+2) = 0) leave;
                Cdr[n] := cons3(NIL, NIL, ATOMB);
                n := Cdr[n];
                i := i + 2;
        end
        n := unsave(1);
        return cons(n, UNDEF);
end

var Symb::SYMLEN+2;

symstr(n) do var i;
        i := 0;
        n := Car[n];
        while (n \= NIL) do
                Symb::i := Car[n] >> 8;
                Symb::(i+1) := Car[n] & 0xff;
                i := i + 2;
                n := Cdr[n];
        end
        Symb::i := 0;
        return Symb;
end

findsym(s) do var p;
	p := Symbols;
	while (p \= NIL) do
		if (strequ(s, symstr(Car[p])))
			return Car[p];
		p := Cdr[p];
	end
	return NIL;
end

addsym(s, v) do var n;
	n := findsym(s);
	if (n \= NIL) return n;
	n := strsym(s);
	Symbols := cons(n, Symbols);
	ie (v = SPCL)
		Cdr[n] := cons(n, NIL);
	else
		Cdr[n] := cons(v, NIL);
	return n;
end

lookup(x) do var a ,e;
	e := Env;
	while (e \= NIL) do
		a := Car[e];
		while (a \= NIL) do
			if (caar(a) = x) return cdar(a);
			a := Cdr[a];
		end
		e := Cdr[e];
	end
	return Cdr[x];
end

var	Parens;

decl read(0);

rdlist() do var n, lst, a, badpair;
	badpair := "bad pair";
	Parens := Parens+1;
	lst := cons(NIL, NIL);
	save(lst);
	a := NIL;
	while (1) do
		if (Errflag) return NIL;
		n := read();
		if (n = EOT) return error("missing ')'", UNDEF);
		if (n = DOT) do
			if (a = NIL) return error(badpair, UNDEF);
			n := read();
			Cdr[a] := n;
			if (n = RPAREN \/ read() \= RPAREN)
				return error(badpair, UNDEF);
			unsave(1);
			Parens := Parens-1;
			return lst;
		end
		if (n = RPAREN) leave;
		ie (a = NIL) 
			a := lst;
		else
			a := Cdr[a];
		Car[a] := n;
		Cdr[a] := cons(NIL, NIL);
	end
	Parens := Parens-1;
	if (a \= NIL) Cdr[a] := NIL;
	unsave(1);
	return a = NIL-> NIL: lst;
end

symbolic(c) return c >= 'a' /\ c <= 'z' \/
		   c >= 'A' /\ c <= 'Z' \/
		   c >= '0' /\ c <= '9' \/
		   c = '-'  \/ c = '/';

var	S_apply, S_if, S_ifnot, S_lambda, S_lamstar, S_macro, S_progn,
	S_quote, S_quasiquote, S_unquote, S_unquote_splice, S_setq;

var	S_t, S_binding, S_cons, S_car, S_cdr, S_atom, S_eq, S_gensym,
	S_it, S_suspend, S_gc, S_eofp, S_load, S_rplaca, S_rplacd,
	S_read, S_prin1, S_print, S_error, S_symbols;

rdsym(c) do var s::SYMLEN+1, i;
	i := 0;
	while (symbolic(c)) do
		if (c = '/') c := rdchci();
		ie (SYMLEN = i)
			error("long symbol", UNDEF);
		else if (i < SYMLEN) do
			s::i := c;
			i := i+1;
		end
		c := rdchci();
	end
	s::i := 0;
	Rejected := c;
	if (strequ(s, "nil")) return NIL;
	return addsym(s, UNDEF);
end

syntax(x) error("syntax", x);

quote(q, n) return cons(q, cons(n, NIL));

rdword() do var s::2, n, c;
	s::1 := 0;
	c := rdchci();
	if (\symbolic(c)) syntax(UNDEF);
	n := cons(NIL, NIL);
	save(n);
	while (1) do
		if (c = '/') c := rdchci();
		s::0 := c;
		Car[n] := addsym(s, UNDEF);
		c := rdchci();
		ie (symbolic(c)) do
			Cdr[n] := cons(NIL, NIL);
			n := Cdr[n];
		end
		else
			leave;
	end
	Rejected := c;
	n := unsave(1);
	return n;
end

read() do var c;
	c := rdchci();
	while (1) do
		while (c = ' ' \/ c = '\t' \/ c = '\n' \/ c = '\r') do
			if (Errflag) return NIL;
			c := rdchci();
		end
		if (c \= ';') leave;
		while (c \= '\n' /\ c \= '\r') c := rdch();
	end
	if (c = EOT \/ c = '%') return EOT;
	if (c = '(') do
		return rdlist();
	end
	if (c = '\'') do
		return quote(S_quote, read());
	end
	if (c = '@' \/ c = '`') do
		return quote(S_quasiquote, read());
	end
	if (c = ',') do
		c := rdchci();
		if (c = '@') return quote(S_unquote_splice, read());
		Rejected := c;
		return quote(S_unquote, read());
	end
	if (c = ')') do
		if (\Parens) error("extra paren", UNDEF);
		return RPAREN;
	end
	if (c = '.') do
		if (\Parens) error("free dot", UNDEF);
		return DOT;
	end
	if (symbolic(c)) do
		return rdsym(c);
	end
	if (c = '#') do
		return rdword();
	end
	syntax(UNDEF);
	return UNDEF;
end

print2(n, d) do
	if (d > PRDEPTH) error("print depth", UNDEF);
	if (Errflag) return error("stop", UNDEF);
	ie (n = NIL) do
		pr("nil");
	end
	else ie (n = EOT) do
		pr("*eot*");
	end
	else ie (n >= SPCL \/ Tag::n & ATOMB) do
		pr("*unprintable*");
	end
	else ie (symbolp(n)) do
		pr(symstr(n));
	end
	else do	! List
		if (\atomp(n) /\ Car[n] = S_lamstar) do
			pr("*closure*");
			return;
		end
		pr("(");
		while (n \= NIL) do
			print2(Car[n], d+1);
			n := Cdr[n];
			ie (symbolp(n)) do
				pr(" . ");
				print2(n, d+1);
				n := NIL;
			end
			else if (n \= NIL) do
				pr(" ");
			end
		end
		if (\Errflag) pr(")");
	end
end

print(n) print2(n, 0);

var	Magic;
var	Id;

dowrite(fd, b, k)
	if (t.write(fd, b, k) \= k)
		error("image write error", UNDEF);

suspend(s) do var fd, k, buf::20;
	fd := t.open(s, T3X.OWRITE);
	if (fd < 0) error("suspend", strsym(s));
	k := strlen(Magic);
	t.memcopy(buf, Magic, k);
	buf::(k) := (NNODES >> 8);
	buf::(k+1) := NNODES;
	buf::(k+2) := (Freelist >> 8);
	buf::(k+3) := Freelist;
	buf::(k+4) := (Symbols >> 8);
	buf::(k+5) := Symbols;
	buf::(k+6) := (Id >> 8);
	buf::(k+7) := Id;
	dowrite(fd, buf, k+8);
	dowrite(fd, Car, NNODES * t.bpw());
	dowrite(fd, Cdr, NNODES * t.bpw());
	dowrite(fd, Tag, NNODES);
	t.close(fd);
end

doread(fd, b, k)
	if (t.read(fd, b, k) \= k)
		fatal("image read error");

fasload(s) do var fd, k, n, m, buf::20;
	fd := t.open(s, T3X.OREAD);
	if (fd < 0) return;
	k := strlen(Magic);
	doread(fd, buf, k+8);
	n := buf::k << 8 | buf::(k+1);
	if (n \= NNODES \/ t.memcomp(buf, Magic, k))
		fatal("bad image");
	Freelist := buf::(k+2) << 8 | buf::(k+3);
	Symbols := buf::(k+4) << 8 | buf::(k+5);
	Id := buf::(k+6) << 8 | buf::(k+7);
	doread(fd, Car, NNODES * t.bpw());
	doread(fd, Cdr, NNODES * t.bpw());
	doread(fd, Tag, NNODES);
	t.close(fd);
end

check(x, k0, kn) do var i, a;
	i := 0;
	a := x;
	while (\atomp(a)) do
		i := i+1;
		a := Cdr[a];
	end;
	if (a \= NIL \/ i < k0 \/ (kn \= %1 /\ i > kn))
		syntax(x);
end

var	Loads;
var	Tmpbuf::NLOAD*BUFLEN;

decl eval(1);

load(s) do var fd, ib, ip, ik, in, re;
	if (Loads >= NLOAD) return error("load limit", UNDEF);
	fd := t.open(s, T3X.OREAD);
	if (fd < 0) return error("load", strsym(s));
	in := Input;
	ip := Inp;
	ik := Ink;
	ib := Inbuf;
	re := Rejected;
	Input := fd;
	Inbuf := @Tmpbuf::(Loads*BUFLEN);
	Inp := 0;
	Ink := 0;
	Rejected := EOT;
	Loads := Loads+1;
	while (\Errflag) do
		Acc := read();
		if (Acc = EOT) leave;
		eval(Acc);
	end
	t.close(fd);
	Loads := Loads-1;
	Rejected := re;
	Inbuf := ib;
	Ink := ik;
	Inp := ip;
	Input := in;
end

type(x) error("type", x);

builtin(x) do var s, n;
	ie (S_car = Car[x]) do
		check(x, 2, 2);
		n := cadr(x);
		if (n = NIL) return NIL;
		if (atomp(n)) type(x);
		return Car[n];
	end
	else ie (S_cdr = Car[x]) do
		check(x, 2, 2);
		n := cadr(x);
		if (n = NIL) return NIL;
		if (atomp(n)) type(x);
		return Cdr[n];
	end
	else ie (S_eq = Car[x]) do
		check(x, 3, 3);
		return cadr(x) = caddr(x)-> S_t: NIL;
	end
	else ie (S_atom = Car[x]) do
		check(x, 2, 2);
		return atomp(cadr(x))-> S_t: NIL;
	end
	else ie (S_cons = Car[x]) do
		check(x, 3, 3);
		return cons(cadr(x), caddr(x));
	end
	else ie (S_rplaca = Car[x]) do
		check(x, 3, 3);
		if (atomp(cadr(x))) type(x);
		Car[cadr(x)] := caddr(x);
		return cadr(x);
	end
	else ie (S_rplacd = Car[x]) do
		check(x, 3, 3);
		if (atomp(cadr(x))) type(x);
		Cdr[cadr(x)] := caddr(x);
		return cadr(x);
	end
	else ie (S_gensym = Car[x]) do
		check(x, 1, 1);
		Id := Id+1;
		s := ntoa(Id);
		s := s-1;
		s::0 := 'G';
		return addsym(s, UNDEF);
	end
	else ie (S_eofp = Car[x]) do
		check(x, 2, 2);
		return cadr(x) = EOT-> S_t: NIL;
	end
	else ie (S_read = Car[x]) do
		check(x, 1, 1);
		return read();
	end
	else ie (S_prin1 = Car[x]) do
		check(x, 2, 2);
		print(cadr(x));
		return cadr(x);
	end
	else ie (S_print = Car[x]) do
		check(x, 2, 2);
		print(cadr(x));
		nl();
		return cadr(x);
	end
	else ie (S_binding = Car[x]) do
		check(x, 2, 2);
		n := cadr(x);
		if (\symbolp(n)) type(x);
		n := lookup(n);
		if (Car[n] = UNDEF) return NIL;
		return n;
	end
	else ie (S_load = Car[x]) do
		check(x, 2, 2);
		if (\symbolp(cadr(x))) type(x);
		load(symstr(cadr(x)));
		return S_t;
	end
	else ie (S_error = Car[x]) do
		check(x, 2, 3);
		if (\symbolp(cadr(x))) type(x);
		ie (cddr(x) = NIL)
			error(symstr(cadr(x)), UNDEF);
		else
			error(symstr(cadr(x)), caddr(x));
		return UNDEF;
	end
	else ie (S_gc = Car[x]) do
		check(x, 1, 2);
		if (Cdr[x] \= NIL)
			Verbose_GC := cadr(x) \= NIL;
		return strsym(ntoa(gc()));
	end
	else ie (S_suspend = Car[x]) do
		check(x, 2, 2);
		if (\symbolp(cadr(x))) type(x);
		suspend(symstr(cadr(x)));
		return S_t;
	end
	else ie (S_symbols = Car[x]) do
		check(x, 1, 1);
		return Symbols;
	end
	else do
		syntax(x);
		return UNDEF;
	end
end

cklam(x) do var p;
	check(x, 3, %1);
	p := cadr(x);
	while (\atomp(p)) do
		if (\symbolp(Car[p])) syntax(x);
		p := Cdr[p];
	end
end

specialp(n)
	return	n = S_quote \/
		n = S_if \/
		n = S_progn \/
		n = S_ifnot \/
		n = S_lambda \/
		n = S_lamstar \/
		n = S_apply \/
		n = S_setq \/
		n = S_macro;

struct MODES = MHALT, MEXPR, MLIST, MBETA, MRETN, MAPPL, MPRED,
		MNOTP, MSETQ, MPROG;

special(x, pm) do
	if (S_quote = Car[x]) do
		check(x, 2, 2);
		pm[0] := munsave();
		return cadr(x);
	end
	if (S_if = Car[x]) do
		check(x, 4, 4);
		msave(MPRED);
		pm[0] := MEXPR;
		save(cddr(x));
		return cadr(x);
	end
	if (S_progn = Car[x]) do
		pm[0] := MEXPR;
		if (Cdr[x] = NIL) return NIL;
		if (cddr(x) = NIL) return cadr(x);
		msave(MPROG);
		save(cddr(x));
		return cadr(x);
	end
	if (S_ifnot = Car[x]) do
		check(x, 3, 3);
		msave(MNOTP);
		pm[0] := MEXPR;
		save(caddr(x));
		return cadr(x);
	end
	if (S_lambda = Car[x]) do
		cklam(x);
		pm[0] := munsave();
		return cons(S_lamstar, cons(Env, Cdr[x]));
	end
	if (S_lamstar = Car[x]) do
		! check(x, 3, %1);
		pm[0] := munsave();
		return x;
	end
	if (S_apply = Car[x]) do
		check(x, 3, 3);
		msave(MAPPL);
		pm[0] := MEXPR;
		save(caddr(x));
		save(UNDEF);
		return cadr(x);
	end
	if (S_macro = Car[x]) do
		check(x, 2, 2);
		if (atomp(cadr(x)) \/ caadr(x) \= S_lambda)
			syntax(x);
		cklam(cadr(x));
		pm[0] := munsave();
		return cons(S_macro,
			cons(cons(S_lamstar, cons(Env, cdadr(x))),
			     NIL));
	end
	if (S_setq = Car[x]) do
		check(x, 3, 3);
		if (\symbolp(cadr(x))) syntax(x);
		msave(MSETQ);
		pm[0] := MEXPR;
		save(cadr(x));
		return caddr(x);
	end
	syntax(x);
	return UNDEF;
end

expand(x) do var n, m, p;
	if (atomp(x)) return x;
	if (Car[x] = S_quote) return x;
	n := symbolp(Car[x])-> Car[lookup(Car[x])]: UNDEF;
	if (\atomp(n) /\ Car[n] = S_macro) do
		m := cons(Cdr[x], NIL);
		m := cons(S_quote, m);
		m := cons(m, NIL);
		m := cons(cadr(n), m);
		m := cons(S_apply, m);
		save(m);
		n := eval(m);
		Car[Stack] := n;
		n := expand(n);
		unsave(1);
		return n;
	end
	p := x;
	while (\atomp(p)) p := Cdr[p];
	if (symbolp(p)) return x;
	save(x);
	n := NIL;
	save(n);
	p := x;
	while (p \= NIL) do
		m := expand(Car[p]);
		n := cons(m, n);
		Car[Stack] := n;
		p := Cdr[p];
	end
	n := nrev(unsave(1));
	unsave(1);
	return n;
end

bindargs(v, a) do var e, n;
	e := NIL;
	save(e);
	while (\atomp(v)) do
		if (a = NIL) return error("too few args", Acc);
		n := cons(Car[v], cons(Car[a], NIL));
		e := cons(n, e);
		Car[Stack] := e;
		v := Cdr[v];
		a := Cdr[a];
	end
	ie (symbolp(v)) do
		n := cons(v, cons(a, NIL));
		e := cons(n, e);
		Car[Stack] := e;
	end
	else if (a \= NIL) do
		return error("extra args", Acc);
	end
	Env := cons(e, Env);
	unsave(1);
end

funapp(x) do
	Acc := x;
	if (atomp(Car[x]) \/ caar(x) \= S_lamstar)
		syntax(x);
	ie (Mstack \= NIL /\ Car[Mstack] = MRETN) do
		Env := cadar(Acc);
		bindargs(caddar(Acc), Cdr[Acc]);
	end
	else do
		save(Env);
		Env := cadar(Acc);
		bindargs(caddar(Acc), Cdr[Acc]);
		msave(MRETN);
	end
	return cons(S_progn, cdddar(x));
end

var	Evlev;

eval(x) do var n, m;
	if (Evlev >= EVDEPTH)
		return error("expansion limit", UNDEF);
	Evlev := Evlev + 1;
	Acc := expand(x);
	msave(MHALT);
	m := MEXPR;
	while (\Errflag) do
		ie (m = MEXPR) do
			if (\Pass /\ tty.query() /\ tty.readc() = 3)
				error("stop", UNDEF);
			ie (symbolp(Acc)) do
				n := Car[lookup(Acc)];
				if (n = UNDEF)
					return error("undefined", Acc);
				Acc := n;
				m := munsave();
			end
			else ie (atomp(Acc)) do
				m := munsave();
			end
			else ie (specialp(Car[Acc])) do
				m := MBETA;
			end
			else do
				save(Cdr[Acc]);
				Acc := Car[Acc];
				save(NIL);
				msave(MLIST);
			end
		end
		else ie (m = MLIST) do
			ie (cadr(Stack) = NIL) do
				Acc := nrev(cons(Acc, unsave(1)));
				unsave(1);
				m := MBETA;
			end
			else do
				Car[Stack] := cons(Acc, Car[Stack]);
				Acc := caadr(Stack);
				Car[Cdr[Stack]] := cdadr(Stack);
				msave(MLIST);
				m := MEXPR;
			end
		end
		else ie (m = MBETA) do
			ie (specialp(Car[Acc])) do
				Acc := special(Acc, @m);
			end
			else ie (symbolp(Car[Acc])) do
				Acc := builtin(Acc);
				m := munsave();
			end
			else do
				Acc := funapp(Acc);
				m := MEXPR;
			end
		end
		else ie (m = MRETN) do
			Env := unsave(1);
			m := munsave();
		end
		else ie (m = MAPPL) do
			ie (Car[Stack] = UNDEF) do
				Car[Stack] := Acc;
				Acc := cadr(Stack);
				msave(MAPPL);
				m := MEXPR;
			end
			else do
				n := unsave(1);
				unsave(1);
				Acc := cons(n, Acc);
				m := MBETA;
			end
		end
		else ie (m = MPRED) do
			n := unsave(1);
			ie (Acc = NIL)
				Acc := cadr(n);
			else
				Acc := Car[n];
			m := MEXPR;
		end
		else ie (m = MNOTP) do
			n := unsave(1);
			ie (Acc = NIL) do
				Acc := n;
				m := MEXPR;
			end
			else do
				m := munsave();
			end
		end
		else ie (m = MSETQ) do
			n := unsave(1);
			Car[lookup(n)] := Acc;
			Acc := n;
			m := munsave();
		end
		else ie (m = MPROG) do
			ie (cdar(Stack) = NIL) do
				Acc := Car[unsave(1)];
				m := MEXPR;
			end
			else do
				Acc := caar(Stack);
				Car[Stack] := cdar(Stack);
				msave(MPROG);
				m := MEXPR;
			end
		end
		else if (m = MHALT) do
			Evlev := Evlev - 1;
			return Acc;
		end
	end
	return NIL;
end

reset() do
	Evlev := 0;
	Parens := 0;
	Loads := 0;
	Errflag := 0;
	Acc := NIL;
	Env := NIL;
	Stack := NIL;
	Mstack := NIL;
	Tmpcar := NIL;
	Tmpcdr := NIL;
	Car[Znode] := NIL;
	Cdr[Znode] := NIL;
end

init() do
	if (\Pass) do
		tty.init();
		if (tty.columns() > BUFLEN) do
			clear();
			fatal("screen too wide");
		end
		tty.mode(1);
		clear();
	end
	Magic := "KL22";
	Symbols := NIL;
	Freelist := NIL;
	Znode := NIL;
	Auxb1::0 := 0;
	Auxb2::0 := 0;
	Id := 0;
	flushinp();
	Verbose_GC := 0;
	t.memfill(Tag, 0, NNODES);
	Znode := cons(NIL, NIL);
	S_t := addsym("t", SPCL);
	S_apply := addsym("apply", UNDEF);
	S_if := addsym("if", UNDEF);
	S_ifnot := addsym("ifnot", UNDEF);
	S_lambda := addsym("lambda", UNDEF);
	S_lamstar := addsym("lambda*", UNDEF);
	S_macro := addsym("macro", UNDEF);
	S_progn := addsym("progn", UNDEF);
	S_quote := addsym("quote", UNDEF);
	S_quasiquote := addsym("quasiquote", UNDEF);
	S_unquote := addsym("unquote", UNDEF);
	S_unquote_splice := addsym("unquote-splice", UNDEF);
	S_setq := addsym("setq", UNDEF);
	S_it := addsym("it", UNDEF);
	S_binding := addsym("binding", SPCL);
	S_cons := addsym("cons", SPCL);
	S_car := addsym("car", SPCL);
	S_cdr := addsym("cdr", SPCL);
	S_atom := addsym("atom", SPCL);
	S_eq := addsym("eq", SPCL);
	S_eofp := addsym("eofp", SPCL);
	S_rplaca := addsym("rplaca", SPCL);
	S_rplacd := addsym("rplacd", SPCL);
	S_gensym := addsym("gensym", SPCL);
	S_read := addsym("read", SPCL);
	S_prin1 := addsym("prin1", SPCL);
	S_print := addsym("print", SPCL);
	S_error := addsym("error", SPCL);
	S_load := addsym("load", SPCL);
	S_gc := addsym("gc", SPCL);
	S_suspend := addsym("suspend", SPCL);
	S_symbols := addsym("symbols", SPCL);
	reset();
end

do var a::14, c;
	c := t.getarg(1, a, 14);
	Pass := 0;
	if (c > 0 /\ strequ(a, "-")) Pass := %1;
	init();
	fasload(c>0 /\ \Pass-> a: "klisp");
	pr("KILO LISP "); pr(@Magic::2); nl();
	if (\Pass) do
		pr(ntoa(gc()));
		pr(" NODES"); nl();
	end
	while (1) do
		reset();
		message("");
		pr("*\s");
		Acc := read();
		if (Errflag) loop;
		if (Acc = EOT) leave;
		message("[EVAL]");
		Acc := eval(Acc);
		if (Errflag) loop;
		print(Acc);
		Car[Cdr[S_it]] := Acc;
		nl();
	end
	if (Pass) nl();
	fini(0);
end
